#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import absolute_import

__author__     = "Lluís Vilanova"
__copyright__  = "Copyright 2008-2014, Lluís Vilanova"
__license__    = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__      = "vilanova@ac.upc.edu"


import sciexp2.version


__version_info__ = sciexp2.version.__version_info__
__version__ = sciexp2.version.__version__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Command-style functions to work with `~sciexp2.launchgen`.

The typical workflow to work with `~sciexp2.launchgen` is::

    import sciexp2.launchgen

    launchgen = sciexp2.launchgen.Launchgen()
    launchgen.out = ...
    ...
    launchgen.execute(...)
    ...
    launchgen.params(var1 = sciexp2.launchgen.defer("var2"))
    ...
    launchgen.pack(...)
    ...
    launchgen.find_files(...)
    ...
    launchgen.launcher(...)

Using this module provides a simplified command-style environment for such a
common workflow, where a `~sciexp2.launchgen.Launchgen` object has been
instantiated as ``default_launchgen`` and all its methods (together with the
`~sciexp2.launchgen.defer` routine) have been put in the namespace::

    from sciexp2.launchgen.env import *

    default_launchgen.out = ...
    ...
    execute(...)
    ...
    params(var1 = defer("var2"))
    ...
    pack(...)
    ...
    find_files(...)
    ...
    launcher(...)

The helper function `~sciexp2.launcher.file_contents` is also provided,
together with `~sciexp2.common.filter.Filter` and an instance of
`~sciexp2.common.filter.PFilter` named `v_` to help streamlining the writing of
filters.

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2013, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


__all__ = []


def _env_add(name, value):
    __all__.append(name)
    globals()[name] = value


def _env_setup():
    import sciexp2.launchgen
    import sciexp2.common.filter

    _env_add("Launchgen", sciexp2.launchgen.Launchgen)
    default_launchgen = sciexp2.launchgen.Launchgen()
    _env_add("default_launchgen", default_launchgen)
    for m in dir(default_launchgen):
        if m[0] != "_":
            _env_add(m, getattr(default_launchgen, m))
    _env_add("file_contents", sciexp2.launchgen.file_contents)
    _env_add("defer", sciexp2.launchgen.defer)
    _env_add("Filter", sciexp2.common.filter.Filter)
    _env_add("v_", sciexp2.common.filter.PFilter())

_env_setup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Relevant functions and objects to work with `~sciexp2.data`.

Provides all "public" contents of the following modules:

* `sciexp2.data`
* `sciexp2.data.io`
* `sciexp2.data.meta`

Also provides `~sciexp2.common.filter.Filter` and an instance of
`~sciexp2.common.filter.PFilter` named `v_` to help streamlining the writing of
filters.

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2013-2014, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


__all__ = []


def _env_add(name, value):
    __all__.append(name)
    globals()[name] = value


def _env_add_all(module):
    for i in module.__all__:
        _env_add(i, getattr(module, i))


def _env_setup():
    import sciexp2.data
    import sciexp2.data.meta
    import sciexp2.data.io
    import sciexp2.common.filter
    _env_add_all(sciexp2.data)
    _env_add_all(sciexp2.data.meta)
    _env_add_all(sciexp2.data.io)
    _env_add("Filter", sciexp2.common.filter.Filter)
    _env_add("v_", sciexp2.common.filter.PFilter())

_env_setup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Common functionality for launcher templates.

Two files can be provided for each template:

- **template-name.tpl**

  The template itself.

  This is the file that will be produced (through :term:`translation`).

- **template-name.dsc**

  The template descriptor.

  Contains a predefined set of attributes (see `Template`).

  The use of the ``parent`` attribute is encouraged to write template
  refinements, so that instead of writing new ``.dsc`` and ``.tpl`` files you
  can simply write a single ``.dsc`` and declare its ``parent``. Then, any
  attribute in the ``.dsc`` file (other than the set of predefined attributes)
  will be used to refine the variables of the parent template.

The templates are searched for in the paths listed in :data:`SEARCH_PATH`,
where you can add new directories if necessary.

.. autodata:: sciexp2.templates.SEARCH_PATH
"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2013, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import glob
import os
import six
import pydoc

from .common import varref
import sciexp2.system


_DICTXLATOR = varref.DictXlator("@")
_TOKENIZER = varref.Tokenizer("@")

#: Paths to search for available templates.
#:
#: The order of the list establishes which template will be used in case it
#: exists in more than one directory.
#:
#: Includes the current directory and the `templates` directory shipped with
#: SciExp².
SEARCH_PATH = [
    os.curdir,
    os.sep.join([os.path.dirname(__file__), "templates"]),
    ]


def _get_path(name):
    """Get the path to descriptor or template file with given name."""
    for path in SEARCH_PATH:
        file_path = os.sep.join([path, name])
        if os.path.isfile(file_path):
            return file_path
    return None


_DOC = pydoc.TextDoc()


def _bold(text):
    return _DOC.bold(text)


def _indent(text):
    return _DOC.indent(text)


class _FakeDict (dict):
    def __getitem__(self, key):
        if key not in self:
            dict.__setitem__(self, key, "")
        return dict.__getitem__(self, key)


class TemplateError (Exception):
    """Error retrieving template file."""
    def __init__(self, message):
        Exception.__init__(self, message)


class Template:
    """A launcher template.

    The non-string attributes must be specified in the template descriptor file
    as the string identifying the object name.

    Attributes
    ----------
    name : str
        Template name (taken from the file name of descriptor file).
    description : str
        Template description.
    system : `sciexp2.system.System`, optional
        Name of the execution system this template if for. Can be inherited
        from `parent`.
    template : str, optional
        The template file this descriptor uses. For file ``name.tpl`` you must
        use ``name``. Can be inherited from `parent`.
    parent : `sciexp2.templates.Template`, optional
        Parent template to inherit from.
    submit_args : list of str, optional
        Extra arguments passed to the job submission program.
    overrides : dict, optional
        Dictionary mapping variable names in the parent template to their
        corresponding values for this template.
    defaults : dict, optional
        Dictionary mapping variable names in the template to their
        corresponding values, in case the user provides none.

    Notes
    -----
    Template descriptor files can use the following variables to refer to the
    corresponding attributes set in their parent template:

    - `parent_submit_args`
    - `parent_overrides`
    - `parent_defaults`

    Argument `submit_args` can contain variable *LAUNCHER_BASE*, which contains
    the absolute path to the base output directory (`launchgen.Launchgen.out`).

    """

    _LOADING = []

    def __init__(self, name):
        """Create a new template object from its descriptor file."""

        if name in Template._LOADING:
            raise TemplateError("Circular template dependency: %s"
                                % " -> ".join(Template._LOADING + [name]))
        self.name = name

        # load descriptor file
        dsc_path = _get_path(self.name + ".dsc")
        if dsc_path is None:
            raise TemplateError("Cannot find template descriptor file '%s'" %
                                self.name + ".dsc")
        # load to know who's the parent
        globals_ = dict(
            parent_submit_args=[],
            parent_overrides=_FakeDict(),
            parent_defaults=_FakeDict(),
        )
        namespace = {}
        execfile(dsc_path, globals_, namespace)

        self._init_parent(namespace, dsc_path)

        # reload with parent-specific information
        if self.parent is None:
            globals_ = dict(
                parent_submit_args=[],
                parent_overrides={},
                parent_default={},
            )
        else:
            globals_ = dict(
                parent_submit_args=list(self.parent.submit_args),
                parent_overrides=dict(self.parent.overrides),
                parent_default=dict(self.parent.defaults),
            )
        namespace = {}
        execfile(dsc_path, globals_, namespace)

        namespace.pop("parent", None)
        self._init_description(namespace, dsc_path)
        self._init_system(namespace, dsc_path)
        self._init_template(namespace, dsc_path)
        self._init_submit_args(namespace)
        self._init_overrides(namespace)
        self._init_defaults(namespace)

        # do not accept any other variable in the descriptor
        if len(namespace) > 0:
            raise TemplateError("Unknown variables in template %s: %s" %
                                (dsc_path, ", ".join(namespace)))

    def _init_parent(self, namespace, dsc_path):
        parent_name = namespace.pop("parent", None)
        if parent_name is not None:
            Template._LOADING.append(self.name)
            try:
                self.parent = get(parent_name)
            except TemplateError as e:
                raise TemplateError("When loading parent of %s: %s" %
                                    (dsc_path, e.message))
            Template._LOADING.remove(self.name)
        else:
            self.parent = None

    def _init_description(self, namespace, dsc_path):
        self.description = namespace.pop("description", None)
        if self.description is None:
            raise TemplateError("Template descriptor without 'description': "
                                "%s" % dsc_path)

    def _init_system(self, namespace, dsc_path):
        self.system = None
        if self.parent is not None:
            self.system = self.parent.system

        system_name = namespace.pop("system", None)
        if system_name:
            try:
                self.system = sciexp2.system.get(system_name)
            except sciexp2.system.SystemError as e:
                raise TemplateError("Error loading 'system' for template "
                                    "%s: %s" % (dsc_path, e.message))
        elif self.system is None:
            raise TemplateError("Template descriptor without 'system': "
                                "%s" % dsc_path)

    def _init_template(self, namespace, dsc_path):
        self.template = None
        if self.parent is not None:
            self.template = self.parent.template
            self.template_path = self.parent.template_path

        template_name = namespace.pop("template", None)
        if template_name:
            self.template = template_name
            self.template_path = _get_path(self.template + ".tpl")
            if self.template_path is None:
                raise TemplateError("Template descriptor with incorrect "
                                    "'template' %r: %s" %
                                    (self.template, dsc_path))
        elif self.template is None:
            raise TemplateError("Template descriptor without 'template': "
                                "%s" % dsc_path)

    def _init_submit_args(self, namespace):
        parent_submit_args = self.parent.submit_args if self.parent else []
        self.submit_args = namespace.pop("submit_args", parent_submit_args)

    def _init_overrides(self, namespace):
        parent_overrides = self.parent.overrides if self.parent else {}
        self_overrides = namespace.pop("overrides", {})
        self.overrides = dict(self_overrides)
        for key, val in parent_overrides.items():
            new_val = _DICTXLATOR.xlate_rec(val, self_overrides,
                                            allow_unxlated=True)
            if new_val != val or key not in self_overrides:
                self.overrides[key] = new_val

    def _init_defaults(self, namespace):
        if self.parent:
            self.defaults = dict(self.parent.defaults)
        else:
            self.defaults = {}
        self.defaults.update(namespace.pop("defaults", {}))

    def get_short_description(self, get_long=False):
        """Get short description."""
        res = [_bold(self.name)]
        contents = []
        contents += [self.description.strip()]
        has_parent = self.parent is not None
        if has_parent:
            contents += [_bold("Parent  : ") + self.parent.name]
        if get_long or not has_parent:
            contents += [_bold("System  : ") + self.system.name]
            contents += [_bold("Template: ") + self.template]
        res.append(_indent("\n".join(contents)))
        return "\n".join(res)

    def get_description(self):
        """Get a full description."""
        res = [self.get_short_description(True)]
        contents = []

        if len(self.submit_args) > 0:
            contents += [_bold("Submit arguments:")]
            contents += [_indent(" ".join(self.submit_args))]
        if len(self.defaults) > 0:
            contents += [_bold("Default values:")]
            defaults = ["%-15s :: \"%s\"" % (var, val)
                        for var, val in sorted(six.iteritems(self.defaults))]
            contents += [_indent("\n".join(defaults))]

        with open(self.template_path) as contents_file:
            mandatory_vars = set([
                v for v in _TOKENIZER.vars(contents_file.read())])
            mandatory_vars |= set([
                v
                for val in six.itervalues(self.overrides)
                for v in _TOKENIZER.vars(val)])
            mandatory_vars -= self.system.assumes()
            mandatory_vars -= self.system.defines()
            mandatory_vars -= set(self.defaults)
            if len(mandatory_vars) > 0:
                contents += [_bold("Mandatory variables:")]
                mandatory = sorted([str(var) for var in mandatory_vars])
                contents += [_indent("\n".join(mandatory))]

        with open(self.template_path) as contents_file:
            contents += [_bold("Contents:")]
            fcontents = "".join(contents_file.readlines())
            fcontents = _DICTXLATOR.xlate(fcontents, self.overrides,
                                          allow_unxlated=True)
            contents += [_indent(fcontents)]

        res += [_indent("\n".join(contents))]
        return "\n".join(res)


_TEMPLATES = {}


def get(name):
    """Get a Template object by name."""
    if name in _TEMPLATES:
        res = _TEMPLATES[name]
    else:
        res = Template(name)
        _TEMPLATES[name] = res
    return res


def _get_all_templates():
    """Search for all possible template file descriptors."""
    for path in SEARCH_PATH:
        for file_path in glob.iglob(path + os.sep + "*.dsc"):
            name = os.path.basename(file_path)[:-4]
            get(name)


def get_description():
    """Get a short description of all available templates."""
    _get_all_templates()
    return "\n\n".join([tpl.get_short_description()
                        for tpl in six.itervalues(_TEMPLATES)])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Find files and extract its contents into a `Data` object.

If files contain the ``.gz`` suffix, the system assumes they are
gzip-compressed.

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2014, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import re
import os
import collections
import functools
import numpy
from six.moves import zip
import six
import sys
import warnings

from sciexp2.common import progress
from sciexp2.common import utils
from sciexp2.common import varref
import sciexp2.common.filter
import sciexp2.common.instance
import sciexp2.common.pp
import sciexp2.data
import sciexp2.launcher
import sciexp2.system
from . import _lazy
from ._errors import *
from . import _source


###########################################################################
## find_files

def _find_files(expression, *filters):
    res = sciexp2.common.instance.InstanceGroup()

    # normalize `filters` format
    new_filters = []
    for f in filters:
        if isinstance(f, collections.Iterable) and \
           not isinstance(f, six.string_types):
            new_filters += f
        else:
            new_filters.append(f)
    filters = sciexp2.common.filter.Filter(*new_filters)

    files = None
    if isinstance(expression, six.string_types) and os.path.isfile(expression):
        try:
            launcher = sciexp2.launcher.load(expression)
            jd = True
            state = sciexp2.system.Job.DONE
            files = launcher._system.build(set([state]), filters)
            with progress.progressable_simple(
                    files, None, msg="finding files...") as pfiles:
                for instance in pfiles:
                    instance = sciexp2.common.instance.Instance(instance)
                    instance["FILE"] = os.sep.join([launcher._system._base,
                                                    instance["DONE"]])
                    for key in ["_STATE", "LAUNCHER", "DONE", "FAIL"]:
                        del instance[key]
                    if filters.match(instance):
                        res += instance
        except sciexp2.launcher.LauncherLoadError:
            pass

    if files is None:
        jd = False
        variables = set(varref.expr_get_vars("@", expression, True)) |\
                    set(["FILE"])
        filters.validate(variables)
        files = utils.find_files("@", expression, "FILE")

        with progress.progressable_simple(
                files, None, msg="finding files...") as pfiles:
            for instance in pfiles:
                if filters.match(instance):
                    res += instance

    return res, jd


def find_files(expression, *filters):
    """Find paths matching a regular expression.

    For each path matching `expression` and `filters`, extract it into an
    `~sciexp2.common.instance.Instance` and add to the result.

    The expression can contain arbitrarily complex `regular expressions
    <http://docs.python.org/library/re.html#regular-expression-syntax>`_,
    and variables are extracted from:

        - variable references: ``@varname@``
        - named regular expression groups: ``(?P<varname>regexp)``


    Parameters
    ----------
    expression : expression
        An expression for selecting paths.
    filters : list of filters, optional
        Filters that paths must match to be considered. All provided filters
        are AND'ed.


    Returns
    -------
    InstanceGroup
        An `~sciexp2.common.instance.InstanceGroup` with the found files.


    Notes
    -----
    Implicitly added variables:
        ======== ======================
        Name     Contents
        ======== ======================
        ``FILE`` the complete path name
        ======== ======================

    If argument *expression* points to a job descriptor file generated by
    `~sciexp2.launchgen.Launchgen.launcher`, the resulting *FILE* values will
    be the path to the *DONE* files for these jobs that have been successfully
    executed.

    For convenience, the elements in *filters* can also be lists of filters.

    The results of multiple calls to `find_files` can be aggregated by
    extending the contents of an `~sciexp2.common.instance.InstanceGroup`
    object.

    You can also set the ``FILE`` variable to a `file`, `StringIO` or any other
    `io.IOBase` subclass instance by either modifying the result of
    `find_files` or by creating an `~sciexp2.common.instance.InstanceGroup` by
    hand.

    """
    return _find_files(expression, *filters)[0]


def _as_find_files(arg):
    """Treat argument as the result of `find_files` or arguments to call it.

    """
    if isinstance(arg, sciexp2.common.instance.InstanceGroup):
        return arg, False
    elif isinstance(arg, str):
        return _find_files(arg)
    elif isinstance(arg, tuple):
        return _find_files(arg[0], *arg[1:])
    else:
        raise TypeError("Invalid 'source' value: %s" % arg)


###########################################################################
## extract_*

#: Data type for string-like elements
STRING_DTYPE = "|S300"


class _FieldInfo (sciexp2.common.pp.Pretty):
    """Descriptor of a field in a source file.

    Parameters
    ----------
    name : str
        Field name.
    index : int
        Field number.
    type
        Field type.
    converter : optional
        Field converter (from string). Default is `type`.
    default : optional
        Default value when field is not present. Default is the default
        constructor of `type`.
    selected : bool, optional
        Whether the field has been selected for extraction. Default is `False`.

    """
    def __init__(self, name, index, type,
                 converter=None, default=None, selected=False):
        self.name = name
        self.index = index
        self.field_name = "f%d" % index
        self.type = type
        if converter is None:
            self.converter = type
        else:
            self.converter = converter
        if default is None:
            self.default = self.type()
        else:
            self.default = default
        self.selected = selected

    def copy(self):
        return _FieldInfo(name=self.name, index=self.index, type=self.type,
                          converter=self.converter, default=self.default,
                          selected=self.selected)

    def _repr_pretty_(self, p, cycle):
        names = ["name", "index", "field_name", "type", "converter", "default",
                 "selected"]
        with self.pformat(p, cycle):
            for name in names[:-1]:
                p.text("%s=%r," % (name, getattr(self, name)))
                p.breakable()
            name = names[-1]
            p.text("%s=%r" % (name, getattr(self, name)))

    def __repr__(self):
        return sciexp2.common.pp.Pretty.__repr__(self)


class _SourceInfo (sciexp2.common.pp.Pretty):
    """Descriptor of the fields in a source file.

    Parameters
    ----------
    fields : list or _SourceInfo
        List of fields (instances of `_FieldInfo`).

    """
    def __init__(self, fields):
        if isinstance(fields, _SourceInfo):
            self._fields = [field.copy() for field in fields]
        else:
            for field in fields:
                if not isinstance(field, _FieldInfo):
                    raise TypeError("Not a `_FieldInfo`: %r" % field)
            self._fields = fields

    def where(self, attribute, value):
        return [field
                for field in self
                if getattr(field, attribute) == value]

    def where_one(self, attribute, value):
        fields = self.where(attribute, value)
        assert len(fields) == 1
        return fields[0]

    def __len__(self):
        return len(self._fields)

    def __iter__(self):
        return iter(self._fields)

    def __getitem__(self, key):
        return self._fields[key]

    def __setitem__(self, key, value):
        self._fields[key] = value

    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            p.pretty(self._fields)

    def __repr__(self):
        return sciexp2.common.pp.Pretty.__repr__(self)


def _probe_fsv_file(fobj, header, delimiter, genfromtxt_args, info=None):
    """Read FSV-formatted `fobj` and return its configuration.

    If `info` is given, returns new information metadata with updated `type`
    and `default` fields according to the `converter` field.
    """
    if info is not None:
        info = _SourceInfo(info)
        info_usecols = [i.index
                        for i in info if i.selected]
        info_converters = dict([
            (i.index, i.converter)
            for i in info
            if i.selected and i.converter is not None])
        if header is None:
            header = [i.name if i.name is not None else i.field_name
                      for i in info]
    else:
        info_usecols = None
        info_converters = None

    with _source.current_position(fobj):
        data = numpy.genfromtxt(fobj, delimiter=delimiter,
                                dtype=None,
                                names=header,
                                usecols=info_usecols,
                                converters=info_converters,
                                **genfromtxt_args)

    if len(data.dtype) > 0:
        names = data.dtype.names
        types = [data.dtype.fields[n][0].type for n in names]
    else:
        if info_usecols is None:
            names = ["f0"]
        else:
            names = [info[i].name for i in info_usecols]
        types = [data.dtype.type]

    # Get options affecting data extraction
    genfromtxt_data_args = {}
    for name in ["missing_values", "filling_values", "invalid_raise"]:
        if name in genfromtxt_args:
            genfromtxt_data_args[name] = genfromtxt_args[name]

    default_contents = six.StringIO(("," * len(names)) + "1")
    dtype = [(n, t) for n, t in zip(names, types)]
    default_data = numpy.genfromtxt(default_contents,
                                    usecols=list(range(len(names))) + [-1],
                                    delimiter=",",
                                    dtype=dtype + [("last", int)],
                                    **genfromtxt_data_args)
    defaults = default_data.tolist()[:-1]

    if info is None:
        info = _SourceInfo([
            _FieldInfo(name=n,
                       index=i,
                       type=t,
                       converter=t,
                       default=d,
                       selected=False)
            for i, (n, t, d) in enumerate(zip(names, types, defaults))
        ])
    else:
        for i in info:
            if not i.selected:
                continue
            didx = info_usecols.index(i.index)
            j = info.where_one("index", didx)
            j.type = types[didx]
            j.default = defaults[didx]

    return info


###########################################################################
### extract_func

def _extract_parse_source(source):
    source_, jd = _as_find_files(source)
    if len(source_) == 0:
        raise ValueError("Cannot accept 'source' with no contents")

    if not isinstance(source, six.string_types) or jd:
        source_vars = list(source_.variables())
        del source_vars[source_vars.index("FILE")]
    else:
        source_vars = varref.expr_get_vars("@", source, True)
    source = source_
    return source, source_vars


def _extract_parse_count(source, kwargs):
    count = kwargs.pop("count", None)
    if count is not None:
        if not isinstance(count, six.string_types):
            raise TypeError("Argument 'count' must be a string")
        if count in source:
            raise ValueError("Variable in 'count' already present in 'source'")
    return count


def _extract_parse_fields_to_vars(kwargs):
    res = kwargs.pop("fields_to_vars", [])
    if not isinstance(res, collections.Iterable):
        raise TypeError("Argument `fields_to_vars` must be a list of "
                        "field names or regular expressions")
    if isinstance(res, six.string_types):
        res = [res]
    res = list(res)
    if not all([isinstance(elem, six.string_types) for elem in res]):
        raise TypeError("Argument `fields_to_vars` must be a list of "
                        "field names or regular expressions")
    return res


def _extract_parse_vars_to_fields(kwargs, source, source_vars):
    names = kwargs.pop("vars_to_fields", [])
    if not isinstance(names, collections.Iterable):
        raise TypeError("Argument `vars_to_fields` must be a list of "
                        "variable names or regular expressions")
    if isinstance(names, six.string_types):
        names = [names]
    names = list(names)
    if not all([isinstance(elem, six.string_types) for elem in names]):
        raise TypeError("Argument `vars_to_fields` must be a list of "
                        "variable names or regular expressions")
    v2f = []
    for v2f_re in names:
        match = False
        v2f_cre = re.compile(v2f_re)
        for source_var in source.variables():
            if v2f_cre.match(source_var) is not None:
                match = True
                if source_var not in v2f:
                    v2f.append(source_var)
        if not match:
            raise ValueError("Value in `vars_to_fields` does not match "
                             "any variable: %s" % v2f_re)
    if len(v2f) == 0:
        return [], source_vars
    source_vars = [var for var in source_vars if var not in v2f]
    if len(source_vars) == 0:
        raise ValueError("Cannot specify all variables in `vars_to_fields`")
    return v2f, source_vars


_EXTRACT_FUNC_ARGS = ["probe_file", "count", "fields_to_vars",
                      "vars_to_fields", "allow_empty"]


def extract_func(source, func, *args, **kwargs):
    r"""Extract data using the provided function.

    Returns a 1-dimensional structured `~sciexp2.data.Data` array, with the
    results of extracting the data of each file in `source`. The result will
    have as many fields as the data produced by `func` (must be the same for
    each source), and as many elements as produced by `func` on each source.

    The variables in `source` (and their values) will be used as the dimension
    variables for the resulting array.


    Parameters
    ----------
    source : str or tuple or result of `find_files`
        `InstanceGroup` as returned by `find_files` with sources to extract
        data from. Can also be a tuple with arguments to `find_files`, or a
        single string with the first argument to `find_files`.
    func : callable
        Function to extract data from a single source file.
    probe_file : optional
        First file to extract data from. Accepts any value valid in the
        ``FILE`` key of `source`. If not in `source`, will only be used to
        probe for data types, but the actual data is ignored. Defaults to the
        first element of `source`.
    count : str, optional
        Dimension variable added to uniquely identify each per-source element
        by its index.
    fields_to_vars : list of str, optional
        Names of fields in the array returned by `func` that will be instead
        used as dimension variables (field values will be used as dimension
        tick values). You can also use regular expressions to select the
        matching fields.
    vars_to_fields : list of str, optional
        Names of variables in `source` that will be instead used as fields in
        the result (variable values will be used as field values). You can also
        use regular expressions to select the matching variables.
    allow_empty : bool, optional
        Whether to allow empty source files, which are ignored (default is
        `True`).
    args, kwargs
        Arguments passed to `func`.


    Raises
    ------
    ProbeError
        An error happened while probing the first source.
    ExtractError
        An error happened while extracting data.
    EmptyFileError
        The first source was empty, or some other was while `allow_empty` was
        `False`.


    Notes
    -----
    If `func` returns an array with one single element for each source, `count`
    and `fields_to_vars` are not necessary. Otherwise, at least one of them is
    required to disambiguate the elements of the resulting array (i.e., cannot
    have two elements with the same dimension ticks).

    Using `fields_to_vars` will drop the selected data fields and instead use
    their values in the dimension metadata.

    Before the actual data extraction, a single source will be selected (the
    first in `source`, or `probe_file` if provided) and used for probing the
    resulting data format (common to all sources). This probed information is
    discarded when later performing the actual data extraction. Thus, probing
    with the first element of `source` will actually extract it twice.

    The provided function `func` must accept the following arguments (plus
    those passed by the user through `args` and `kwargs`):

    `source`
        The file to extract data from (assumes a `file`-like interface).
    `source_first` : bool
        Whether `source` is being used for probing.

    The result of `func` must be a structured `numpy.ndarray` with at least one
    field and one or two dimensions (depending on whether `count` is being
    used).

    """
    ##################################################
    # parse arguments

    source, source_vars = _extract_parse_source(source)
    count = _extract_parse_count(source, kwargs)
    f2v = _extract_parse_fields_to_vars(kwargs)
    v2f, source_vars = _extract_parse_vars_to_fields(
        kwargs, source, source_vars)
    allow_empty = kwargs.pop("allow_empty", True)
    if not callable(func):
        raise TypeError("Argument 'func' is not callable")

    source_expr = "@" + "@-@".join(source_vars) + "@" \
                  if len(source_vars) > 0 else ""

    probe_file = kwargs.pop("probe_file", None)
    if probe_file is None:
        probe_file = source.get_index(0)["FILE"]
        probe_id = _source.get_name(probe_file, 0)
    else:
        probe_id = _source.get_name(probe_file, None)

    ##################################################
    # extract first file and post-process arguments

    def _extract_func_file(source_file, source_id, error_cls, probe,
                           *args, **kwargs):
        if _source.is_empty(source_file):
            if allow_empty:
                return []
            else:
                raise EmptyFileError(source_id)
        try:
            data = func(source=source_file, source_first=probe,
                        *args, **kwargs)
            # sanity checks
            if not isinstance(data, numpy.ndarray):
                raise TypeError("Extraction function returned non-numpy data")
            if len(data.dtype) == 0 or len(data.dtype.names) == 0:
                raise TypeError("Extraction function returned data without fields")
            if data.shape == tuple():
                raise TypeError("Extraction function returned data without shape")

            if len(data) > 1 and count is None and len(f2v) == 0:
                raise ValueError("Source has multiple results, you must use "
                                 "'count' and/or 'fields_to_vars'")
        except EmptyFileError:
            raise
        except Exception as e:
            _, _, exc_traceback = sys.exc_info()
            new_e = error_cls(source_id, e)
            six.reraise(new_e.__class__, new_e, exc_traceback)

        if len(data) == 0 and not allow_empty:
            raise EmptyFileError(source_id)
        return data

    probe_file = _source.get_file(probe_file)
    if _source.is_empty(probe_file):
        raise EmptyFileError(probe_id)
    data = _extract_func_file(probe_file, probe_id, ProbeError, True,
                              *args, **kwargs)

    # find fields in `fields_to_vars`
    if len(f2v) > 0:
        f2v_ = []
        for pidx, p in enumerate(f2v):
            pfound = False
            pre = re.compile(p + "$")
            for idx, name in enumerate(data.dtype.names):
                field_name = "f%d" % idx
                if (pre.match(name) or pre.match(field_name)):
                    f2v_.append(name)
                    pfound = True
            if not pfound:
                raise ValueError(
                    ("Entry number %d (%s) in " % (pidx, p)) +
                    "`fields_to_vars` does not match any extracted variable")
        f2v = f2v_

    # get index of `fields_to_vars` elements
    if len(f2v) > 0:
        def _find_f2v_idx(name):
            for idx, n in enumerate(data.dtype.names):
                if name == n:
                    return idx
        f2v_dict = dict([(p, _find_f2v_idx(p)) for p in f2v])
        f2v_indexes = [f2v_dict[name] for name in f2v]
        f2v_indexes.sort()

    # final sanity checks for arguments
    if count in f2v:
        raise ValueError("Variable in 'count' already present in "
                         "`fields_to_vars`")
    if all(name in f2v for name in data.dtype.names):
        raise ValueError("Cannot move all fields into variables "
                         "(`fields_to_vars`)")

    ##################################################
    # mass-extraction

    def _extract_func_postprocess(data, source, v2f):
        data_dim = []
        data_values = []

        for i, ldata in enumerate(data):
            variables = []
            values = list(ldata.tolist())

            if count is not None:
                variables.append((count, i))
            if len(f2v) > 0:
                for name in f2v:
                    variables.append((name, values[f2v_dict[name]]))
                for pidx in reversed(f2v_indexes):
                    del values[pidx]

            data_dim.append(dict(
                [(key, val) for key, val in source.items()
                 if key not in v2f] + variables))
            data_values.append(tuple(values + [source[var] for var in v2f]))

        return data, data_values, data_dim

    dim = []
    dtype = [(data.dtype.names[i], data.dtype[i])
             for i in range(len(data.dtype))
             if data.dtype.names[i] not in f2v]
    for var in v2f:
        tmp = numpy.array(list(source[var].keys()))
        if numpy.issubdtype(tmp.dtype, str):
            vdtype = STRING_DTYPE
        else:
            vdtype = tmp.dtype
        dtype.append((var, vdtype))
    arr, growable = _arr_growable_init(
        len(source), data, dtype)

    with progress.progressable_simple(
            source, None, msg="extracting files...") as psource:
        for source_idx, source_name in enumerate(psource):
            source_file = _source.get_file(source_name["FILE"])
            source_id = _source.get_name(source_file, source_idx)
            data = _extract_func_file(source_file, source_id, ExtractError,
                                      False, *args, **kwargs)
            if len(data) > 0:
                data, data_values, data_dim = _extract_func_postprocess(
                    data, source_name, v2f)
                arr, growable = _arr_growable_fill(arr, growable,
                                                   len(source), dim,
                                                   data_values, data_dim)

    arr = _arr_growable_fini(arr, dim)

    if len(f2v) > 0:
        if source_expr != "":
            source_expr += "-"
        source_expr += "@" + "@-@".join(f2v) + "@"
    if count:
        if source_expr != "":
            source_expr += "-"
        source_expr += "@%s@" % count
    return sciexp2.data.data_array(arr,
                                   dims=[(source_expr, dim)],
                                   dtype=arr.dtype)


def _arr_growable_init(sources, data, dtype):
    avg_lines = float(len(data))
    avg_count = 1
    arr = numpy.ndarray(sources * int(avg_lines), dtype=dtype)
    return arr, (avg_lines, avg_count)


def _arr_growable_fill(arr, info, sources, dim, arr_data, dim_data):
    """Fill-in contents and grow if necessary."""
    assert len(arr_data) == len(dim_data)
    avg_lines, avg_count = info
    avg_lines = (len(arr_data[0]) * len(arr_data)) + (avg_count * avg_lines)
    avg_lines = avg_lines / (avg_count + 1)
    avg_count += len(arr_data)

    # grow array, accounting for averages
    if len(dim) + len(dim_data) > len(arr):
        size = max(sources * int(avg_lines),
                   len(dim) + len(dim_data))
        arr_ = numpy.ndarray(size, dtype=arr.dtype)
        arr_[:len(arr)] = arr
        arr = arr_

    arr[len(dim):len(dim)+len(dim_data)] = arr_data
    dim.extend(dim_data)

    return arr, (avg_lines, avg_count)


def _arr_growable_fini(arr, dim):
    """Trims excess elements."""
    if len(dim) < len(arr):
        arr = arr[:len(dim)]
    return arr


###########################################################################
### extract_txt

def _txt_normalize_descriptor(index, descriptor):
    error = False
    fields = {"regex": None, "name": None, "converter": None}

    if isinstance(descriptor, six.string_types):
        fields["regex"] = descriptor
    elif callable(descriptor):
        fields["converter"] = descriptor
    elif isinstance(descriptor, collections.Iterable):
        if len(descriptor) == 2:
            if isinstance(descriptor[0], six.string_types) and \
               isinstance(descriptor[1], six.string_types):
                fields["regex"] = descriptor[0]
                fields["name"] = descriptor[1]
            elif isinstance(descriptor[0], six.string_types) and \
                 callable(descriptor[1]):
                fields["regex"] = descriptor[0]
                fields["converter"] = descriptor[1]
            else:
                error = True
        elif len(descriptor) == 3:
            if isinstance(descriptor[0], six.string_types) and \
               isinstance(descriptor[1], six.string_types) and \
               callable(descriptor[2]):
                fields["regex"] = descriptor[0]
                fields["name"] = descriptor[1]
                fields["converter"] = descriptor[2]
            else:
                error = True
        else:
            error = True
    else:
        error = True

    if error:
        msg = "Invalid element in position %d " % index
        msg += "of `fields`: %s" % repr(descriptor)
        raise ValueError(msg)

    if fields["regex"] is not None:
        fields["regex"] = re.compile(fields["regex"] + "$")

    return fields


def extract_txt(source, *fields, **kwargs):
    r"""Extract data from structured text files.

    Uses `extract_func` to extract data from each file in `source` and returns a
    1-dimensional structured `~sciexp2.data.Data` array. Sources must be in a
    field-separated-value format (e.g., CSV with one column per field). For each
    source, the array has as many fields as selected by `fields`, and as many
    elements as lines.

    If any data element is missing, its default value is set according to
    `numpy.genfromtxt`.


    Parameters
    ----------
    source
        Argument to `extract_func`.
    fields : optional
        List describing which and how to extract fields from the source files
        (by default extracts all fields and auto-detects their format).
    delimiter : str, optional
        The string used to separate values (default is ``,``, which corresponds
        to CSV format).
    header : bool, optional
        Whether the source files contain a header describing the fields.
        The header is used to provide a field name, which otherwise defaults
        to ``f<num>`` (default is `True`).
    probe_file, count, fields_to_vars, allow_empty : optional
        Arguments to `extract_func`.
    comments, skip_header, skip_footer, missing_values, filling_values,
    excludelist, deletechars, defaultfmt, autostrip, replace_space,
    case_sensitive, invalid_raise : optional
        Arguments passed to `numpy.genfromtxt`.


    Returns
    -------
    Data
        A 1-dimensional `~sciexp2.data.Data` array with the extracted data.


    Notes
    -----
    Elements in argument `fields` have the base format ``(regexp, name, type or
    func)``, where each is:

    * Regular expression to match on each field name.

      Can be applied to both the source header (if available) and the default
      field name (see `defaultfmt`).
      If present, only fields matched by any regexp will be extracted.

    * New field name.

      Used to rename a field, can reference groups in the previous regular
      expression.

    * Field value type or function to convert the string in the source file to
      its actual value.

      If the converter raises an exception, a default value will be assigned
      according to the field's type.

    A few shortcuts are also available:

        ==========================  ===========================================
        Shortcut                    Description
        ==========================  ===========================================
        ``regexp``                  Select matching fields.
        ``type or func``            Use specified converter on selected field
                                    (selects all fields if none was).
        ``(regexp, name)``          Select matching fields and rename.
        ``(regexp, type or func)``  Select matching fields and convert.
        ==========================  ===========================================

    Since `fields` is a list, the last matching element on each field takes
    precedence on the selected action. Thus the order in `fields` can be used
    to refine the actions.

    When used, `fields_to_vars` is applied after renaming fields.


    Examples
    --------
    >>> cat data/extract_txt/foo.csv    # doctest: +SKIP
    time,iterations,test
    12.3,5,fast
    53.8,9,slow
    ,,failed
    >>> extract_txt("data/extract_txt/@configuration@.csv",
    ...             count="line")
    data_array([(8.6, 2, 'fast'),
                (53.36, 1, 'slow'),
                (nan, -1, 'failed'),
                (12.3, 5, 'fast'),
                (53.8, 9, 'slow'),
                (nan, -1, 'failed')],
               dtype=[('time', '<f8'), ('iterations', '<i8'), ('test', '|S300')],
               dims=[('@configuration@-@line@',
                      ['bar-0', 'bar-1', 'bar-2', 'foo-0', 'foo-1', 'foo-2'])])

    You can limit which files are extracted without explicitly using
    `find_files`:

    >>> extract_txt(("data/extract_txt/@configuration@.csv",
    ...              "configuration == 'bar'"),
    ...             count="line")
    data_array([(8.6, 2, 'fast'), (53.36, 1, 'slow'), (nan, -1, 'failed')],
               dtype=[('time', '<f8'), ('iterations', '<i8'), ('test', '|S300')],
               dims=[('@configuration@-@line@', ['bar-0', 'bar-1', 'bar-2'])])

    Using `fields_to_vars` is usually more convenient than `count` when
    discriminating between multiple elements:

    >>> extract_txt("data/extract_txt/@configuration@.csv",
    ...             fields_to_vars=["test"])
    data_array([(8.6, 2), (53.36, 1), (nan, -1), (12.3, 5), (53.8, 9), (nan, -1)],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@configuration@-@test@',
                      ['bar-fast',
                       'bar-slow',
                       'bar-failed',
                       'foo-fast',
                       'foo-slow',
                       'foo-failed'])])

    You can use Python's support for regular expressions to rename fields, as
    well as use your own functions to convert field values:

    >>> def str_float2int(n): return int(float(n))
    >>> extract_txt("data/extract_txt/@configuration@.csv",
    ...             ("(.*)", r"Col_\1"),
    ...             ("time", str_float2int),
    ...             count="line")
    data_array([(8, 2, 'fast'),
                (53, 1, 'slow'),
                (-1, -1, 'failed'),
                (12, 5, 'fast'),
                (53, 9, 'slow'),
                (-1, -1, 'failed')],
               dtype=[('Col_time', '<i8'),
                      ('Col_iterations', '<i8'),
                      ('Col_test', '|S300')],
               dims=[('@configuration@-@line@',
                      ['bar-0', 'bar-1', 'bar-2', 'foo-0', 'foo-1', 'foo-2'])])

    See also
    --------
    extract_func
    re

    """
    delimiter = kwargs.pop("delimiter", ",")

    header = kwargs.pop("header", True)
    header = True if header else None

    # Arguments specific to genfromtxt
    genfromtxt_args = {}
    for name in [
            "comments", "skip_header", "skip_footer", "missing_values",
            "filling_values", "excludelist", "deletechars", "defaultfmt",
            "autostrip", "replace_space", "case_sensitive", "invalid_raise"
    ]:
        if name in kwargs:
            genfromtxt_args[name] = kwargs.pop(name)

    kwargs_unhandled = set(kwargs) - set(_EXTRACT_FUNC_ARGS)
    if len(kwargs_unhandled) > 0:
        raise TypeError("Unexpected arguments %s" %
                        ", ".join(kwargs_unhandled))

    state = {}
    return extract_func(source, _extract_txt_func,
                        fields=fields,
                        delimiter=delimiter,
                        header=header,
                        genfromtxt_args=genfromtxt_args,
                        state=state,
                        **kwargs)


def _extract_txt_func(source, source_first, header, delimiter, fields,
                      genfromtxt_args, state):
    if source_first:
        return _extract_txt_probe(source,
                                  fields=fields,
                                  delimiter=delimiter,
                                  header=header,
                                  genfromtxt_args=genfromtxt_args,
                                  state=state)
    else:
        return _extract_txt_file(source,
                                 delimiter=delimiter,
                                 genfromtxt_args=genfromtxt_args,
                                 **state)


def _extract_txt_probe(source, fields, delimiter, header, genfromtxt_args,
                       state):
    ##################################################
    # parse field information and transform according to `fields`

    info_orig = _probe_fsv_file(source,
                                header=header,
                                delimiter=delimiter,
                                genfromtxt_args=genfromtxt_args)
    info = _SourceInfo(info_orig)

    for index, descriptor in enumerate(fields):
        # normalize descriptors
        fields = _txt_normalize_descriptor(index, descriptor)

        # modify fields in info_orig with descriptors
        if fields["regex"] is None:
            assert fields["name"] is None
            if index == 0:
                # select all
                for i in info:
                    i.selected = True
            # apply on selected
            for i in info:
                if i.selected and fields["converter"] is not None:
                    i.converter = fields["converter"]
        else:
            # apply on regexp
            for i in info_orig:
                match = fields["regex"].match(i.name)
                if match is None:
                    match = fields["regex"].match(i.field_name)

                if match is not None:
                    j = info.where_one("field_name", i.field_name)
                    j.selected = True
                    if fields["name"] is not None:
                        j.name = match.expand(fields["name"])
                    if j.selected and fields["converter"] is not None:
                        j.converter = fields["converter"]

    if len(fields) == 0:
        for i in info_orig:
            info.where_one("field_name", i.field_name).selected = True

    ##################################################
    # recalculate types by applying converters

    info = _probe_fsv_file(source, header=header, delimiter=delimiter,
                           genfromtxt_args=genfromtxt_args, info=info)

    ##################################################
    # adjust arguments

    # skip header when extracting actual data
    if header:
        if "skip_header" not in genfromtxt_args:
            genfromtxt_args["skip_header"] = 0
        genfromtxt_args["skip_header"] += 1

    # make strings long enough
    for i in info:
        if issubclass(i.type, six.string_types):
            i.type = STRING_DTYPE

    # arguments for extraction using `genfromtxt`
    state["x_dtype"] = [
        (i.name, i.type)
        for i in info if i.selected]
    state["x_usecols"] = [
        i.index
        for i in info if i.selected]
    state["x_converters"] = dict([
        (i.index, i.converter)
        for i in info
        if i.selected and i.converter is not None])
    if len(state["x_converters"]) == 0:
        state["x_converters"] = None

    ##################################################
    # extract probed file

    return _extract_txt_file(source,
                             delimiter=delimiter,
                             genfromtxt_args=genfromtxt_args,
                             **state)


def _extract_txt_file(source, delimiter, x_usecols, x_converters, x_dtype,
                      genfromtxt_args):
    data = numpy.genfromtxt(source, delimiter=delimiter,
                            usecols=x_usecols,
                            converters=x_converters,
                            dtype=x_dtype,
                            names=None,
                            **genfromtxt_args)

    # single-field source
    if len(data.dtype) == 0:
        if data.shape == tuple():
            data = numpy.array([(data,)], dtype=x_dtype)
        else:
            data = numpy.array([(i,) for i in data], dtype=x_dtype)
    # single-row source
    if data.shape == tuple():
        data = numpy.array([data], dtype=data.dtype)

    return data


###########################################################################
### extract_regex

def _regex_normalize(expression, index):
    """Replace variable references with `re` named groups or backreferences."""
    if isinstance(expression, tuple):
        if len(expression) != 2 or \
           not isinstance(expression[0], six.string_types):
            raise TypeError("Invalid regex entry %d: %s" % (index, expression))
        converter = expression[1]
        expression = expression[0]
        if not isinstance(converter, dict):
            converter = {None: converter}
        converter_default = converter.pop(None, None)
        for var in varref.expr_get_vars("@", expression, accept_regexp=True):
            converter[var] = converter.get(var, converter_default)
        if not all(callable(c) or c is None for c in converter.values()):
            raise TypeError("Invalid regex entry %d: %s" % (index, expression))
    elif not isinstance(expression, six.string_types):
        raise TypeError("Invalid regex entry %d: %s" % (index, expression))
    else:
        converter = {}
        for var in varref.expr_get_vars("@", expression, accept_regexp=True):
            converter[var] = None

    return varref.expr_to_regexp("@", expression), converter


def _regex_extract(source, cre):
    if isinstance(cre, list):           # multi_line
        cres = cre
        contents = "".join(source.readlines())
        values = []
        value = {}
        for cre in cres:
            for i, m in enumerate(cre.finditer(contents)):
                values = values + [{}] * (i - len(values) + 1)
                if m is None:
                    continue
                for k, v in six.iteritems(m.groupdict()):
                    if v is None:
                        continue
                    values[i][k] = v
        return values
    else:
        values = []
        value = {}
        for line in source:
            m = cre.search(line)
            if m is None:
                continue
            if any(v is not None and k in value and value[k] is not None
                   for k, v in six.iteritems(m.groupdict())):
                # new set of results
                values.append(value)
                value = {}
            for k, v in six.iteritems(m.groupdict()):
                if v is None:
                    continue
                value[k] = v
        values.append(value)
        return values


def extract_regex(source, *fields, **kwargs):
    r"""Extract data using regular expressions.

    Uses `extract_func` to extract data from each file in `source` and returns
    a 1-dimensional structured `~sciexp2.data.Data` array. Data in sources is
    extracted using the regular expressions in `fields`. For each source, the
    array has as many fields as regular expressions in `fields`, and as many
    elements as different matches in `fields` (e.g., if two lines match the
    same regular expression, the resulting data will have two elements).

    If any data element is missing, its default value is set according to
    `numpy.genfromtxt`.


    Parameters
    ----------
    source
        Argument to `extract_func`.
    fields
        List describing how to extract data from the source files.
    multi_line : bool, optional
        Whether to allow regexps in `fields` to match multiple lines (default
        is no, which is faster).
    probe_file, count, fields_to_vars, allow_empty : optional
        Arguments to `extract_func`.


    Returns
    -------
    Data
        A 1-dimensional `~sciexp2.data.Data` array with the extracted data.


    Notes
    -----
    Elements in argument `fields` have the base format ``(regexp, dict)``,
    where each is:

    * Regular expression to match on the source contents.

      Variable references in the regular expression will be used to extract a
      value into a field named as the variable.

      In the case of complex patterns, you can use named regular expression
      groups instead of variable references; i.e., ``@var@`` is a shorthand for
      ``(?P<var>.+)``.

    * Dictionary mapping variables (in the regexp) to value types or functions
      used to convert the string in the source file to its actual value.

      Using ``None`` as a key will apply that type or function to convert the
      values of all variables not explicitly listed as keys in the dictionary.

      If no value was extracted (i.e., an optional regexp group), the converter
      will receive `None`.

      If the converter raises an exception, a default value will be assigned
      according to the field's type.

    Every element in `fields` can also be expressed in any of the following
    shortcuts:

    * ``regexp``

      The field type will be inferred automatically from the probing file.

    * ``(regexp, type or func)``

      A shortcut for ``(regexp, {None: type or func})``.


    Examples
    --------
    As an expression can contain multiple variable references, you can extract
    more than one value from every line:

    >>> cat data/extract_regex/foo.txt  # doctest: +SKIP
    test: fast
      time = 12.3 (5 iterations)
      test: complete
    test: slow
      time = 53.8 (9 iterations)
      test: complete
    test: failed
      test: failed
    >>> def str_float2int(n): return int(float(n))
    >>> extract_regex("data/extract_regex/@configuration@.txt",
    ...               "^test: @test@",
    ...               ("time = @time@ \\(@iterations@ ", str_float2int),
    ...               count = "num")
    data_array([('fast', 53, 1),
                ('fast', 12, 5),
                ('slow', 53, 9),
                ('failed', -1, -1)],
               dtype=[('test', '|S300'), ('time', '<i8'), ('iterations', '<i8')],
               dims=[('@configuration@-@num@',
                      ['bar-0', 'foo-0', 'foo-1', 'foo-2'])])

    You can limit which files are extracted without explicitly using
    `find_files`:

    >>> extract_regex(("data/extract_regex/@configuration@.txt",
    ...                "configuration=='bar'"),
    ...               "^test: @test@",
    ...               ("time = @time@ \\(@iterations@ ", str_float2int),
    ...               count = "num")
    data_array([('fast', 53, 1)],
               dtype=[('test', '|S300'), ('time', '<i8'), ('iterations', '<i8')],
               dims=[('@configuration@-@num@', ['bar-0'])])

    Using `fields_to_vars` is usually more convenient than `count` when
    discriminating between multiple elements:

    >>> extract_regex("data/extract_regex/@configuration@.txt",
    ...               "^test: @test@",
    ...               ("time = @time@ \\(@iterations@ ", str_float2int),
    ...               fields_to_vars=["test"])
    data_array([(53, 1), (12, 5), (53, 9), (-1, -1)],
               dtype=[('time', '<i8'), ('iterations', '<i8')],
               dims=[('@configuration@-@test@',
                      ['bar-fast', 'foo-fast', 'foo-slow', 'foo-failed'])])


    See also
    --------
    extract_func
    re

    """
    # `fields`
    if len(fields) == 0:
        raise ValueError("Argument `fields` cannot be empty")
    fields = [_regex_normalize(r, i) for i, r in enumerate(fields)]

    multi_line = kwargs.pop("multi_line", False)

    kwargs_unhandled = set(kwargs) - set(_EXTRACT_FUNC_ARGS)
    if len(kwargs_unhandled) > 0:
        raise TypeError("Unexpected arguments: %s" %
                        ", ".join(kwargs_unhandled))

    converters = {}
    for field in fields:
        convs = field[1]
        for key in convs:
            if key in converters:
                raise ValueError("Duplicate field name: %s" % key)
            converters[key] = convs[key]
    regexes = [i[0] for i in fields]

    state = {}
    return extract_func(source, _extract_regex_func,
                        regexes=regexes,
                        converters=converters,
                        multi_line=multi_line,
                        state=state,
                        **kwargs)


def _extract_regex_func(source, source_first, regexes, converters, multi_line,
                        state):
    if source_first:
        return _extract_regex_probe(source, regexes, converters, multi_line,
                                    state)
    else:
        return _extract_regex_file(source, **state)


def _extract_regex_probe(source, regexes, converters, multi_line, state):
    ##################################################
    # establish data types from inital source

    variables = []
    for regex in regexes:
        regex_vars = varref.regexp_get_names(regex)
        variables.extend(regex_vars)

    if multi_line:
        state["x_cre"] = [re.compile(regex, flags=re.MULTILINE)
                          for regex in regexes]
    else:
        state["x_cre"] = re.compile("(" + ")|(".join(regexes) + ")")

    with _source.current_position(source):
        data = _regex_extract(source, state["x_cre"])
        if len(data) == 0:
            raise EmptyFileError(_source.get_name(source, None))
        data = data[0]
        if len(data) == 0:
            raise EmptyFileError(_source.get_name(source, None))

    data_contents = []
    for var in variables:
        conv = converters[var]
        val = data.get(var, None)
        if conv is not None:
            # if invalid v, conversion can throw an exception (e.g., float(""))
            try:
                val = conv(val)
            except Exception:
                val = ""
            val = str(val)
        elif val is None:
            val = ""
        else:
            # XXX: avoid spurious complaints
            val = val.replace("\n", "")
        data_contents.append(val)
    data_contents = ",".join(data_contents)
    contents = ",".join(variables) + "\n" + data_contents
    # TODO: do not probe non-missing entries with explicit converter
    assert len(contents) > 0
    info = _probe_fsv_file(six.StringIO(contents),
                           header=True,
                           delimiter=",",
                           genfromtxt_args={})

    # apply user-defined converters
    for regex in regexes:
        for var in varref.regexp_get_names(regex):
            conv = converters[var]
            if conv is not None:
                info.where_one("name", var).converter = conv
    for i in info:
        i.selected = True

    ##################################################
    # recalculate types by applying converters

    if len(data_contents) == 0:
        data_contents = ","
    info = _probe_fsv_file(six.StringIO(data_contents),
                           header=None,
                           delimiter=",",
                           genfromtxt_args={},
                           info=info)

    ##################################################
    # adjust arguments

    # make strings long enough
    for i in info:
        if issubclass(i.type, six.string_types):
            i.type = STRING_DTYPE

    # arguments for extraction
    state["x_dtype"] = [(i.name, i.type) for i in info]
    state["x_defaults"] = [i.default for i in info]
    state["x_converters"] = [i.converter for i in info]
    state["x_map"] = dict([(i.name, i.index) for i in info])

    ##################################################
    # extract probed file

    return _extract_regex_file(source, **state)


def _extract_regex_file(source,
                        x_cre, x_dtype, x_defaults, x_converters, x_map):
    data = _regex_extract(source, x_cre)

    variables = [var for var, _ in sorted(x_map.items(),
                                          key=lambda (var, idx): idx)]

    # apply converters, use defaults otherwise
    data_values = []
    for ldata in data:
        values = []
        for idx, (conv, var) in enumerate(zip(x_converters, variables)):
            val = ldata.get(var, None)
            # conversion can throw an exception (e.g., float(""))
            try:
                val = conv(val)
            except Exception:
                val = x_defaults[idx]
            values.append(val)

        data_values.append(tuple(values))

    data = numpy.ndarray(len(data_values), dtype=x_dtype)
    data[:] = data_values

    return data


###########################################################################
### lazy

def lazy(func, *args, **kwargs):
    """Lazily evaluate function with given arguments.

    Returns an object for lazy evaluation, which can be used as a parameter to
    other lazily evaluated functions, or can be used to lazily invoke methods
    on it.

    The `func` argument can be any arbitrary function, including those defined
    by the user.

    The actual result is computed by calling the `realize(path)` method on the
    resulting object. If `path` does not exist, the result will be computed and
    saved to disk. If `path` exists but was generated from older results or
    different arguments, the result is also computed and saved to disk. In the
    common case where it is supposed to compute the exact same result, it
    simply loads the result from disk.

    The `path` argument to `realize` is optional. If it is not provided, it
    will compute the result, but will not save it to disk.

    Similarly, the `checkpoint(path)` method returns a new lazily evaluated
    result that will be saved to disk (just like `realize`).


    Parameters
    ----------
    func
        Function to lazily evaluate.
    args, kwargs
        Arguments to `func`.


    See also
    --------
    lazy_wrap, lazy_wrap_realize, lazy_wrap_checkpoint


    Examples
    --------
    The use of `lazy` has some knowledge about data extraction, and will force
    it only if newer data is available, otherwise loading the previous result
    from "/tmp/extracted":

    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> data = data.realize("/tmp/extracted")
    >>> data
    data_array([(8.6, 2), (53.36, 1), (nan, -1), (12.3, 5), (53.8, 9), (nan, -1)],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@configuration@-@test@',
                      ['bar-fast',
                       'bar-slow',
                       'bar-failed',
                       'foo-fast',
                       'foo-slow',
                       'foo-failed'])])

    For convenience, invoking a method on a lazily evaluated result returns a
    new lazily evaluated result:

    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> data = data.reshape(["configuration"])
    >>> data = data.realize("/tmp/reshaped")

    This is equivalent to:

    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> data = lazy(sciexp2.data.Data.reshape, data, ["configuration"])
    >>> data = data.realize("/tmp/reshaped")
    >>> data
    data_array([[(8.6, 2), (53.36, 1), (nan, -1)],
                [(12.3, 5), (53.8, 9), (nan, -1)]],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@configuration@', ['bar', 'foo']),
                     ('@test@', ['fast', 'slow', 'failed'])])

    If an intermediate result is used by many other functions, you can save it
    to disk with `checkpoint(path)`. In this example, only the calculations for
    `reshape` are reexecuted if "/tmp/extracted" already exists and is up to
    date:

    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> data = data.checkpoint("/tmp/extracted")
    >>> data1 = data.reshape(["configuration"])
    >>> data1 = data1.realize("/tmp/reshaped1")
    >>> data1
    data_array([[(8.6, 2), (53.36, 1), (nan, -1)],
                [(12.3, 5), (53.8, 9), (nan, -1)]],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@configuration@', ['bar', 'foo']),
                     ('@test@', ['fast', 'slow', 'failed'])])
    >>> data2 = data.reshape(["test"])
    >>> data2 = data2.realize("/tmp/reshaped2")
    >>> data2
    data_array([[(8.6, 2), (12.3, 5)],
                [(53.36, 1), (53.8, 9)],
                [(nan, -1), (nan, -1)]],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@test@', ['fast', 'slow', 'failed']),
                     ('@configuration@', ['bar', 'foo'])])

    """
    return _lazy.Result(func, args, kwargs)


def lazy_wrap(func):
    """Convenience decorator/wrapper for lazily-evaluated user functions.

    Invoking ``func(args)`` when `func` has been wrapped/decorated with
    `lazy_wrap` is equivalent to calling ``lazy(func, args)``.


    See also
    --------
    lazy, lazy_wrap_realize, lazy_wrap_checkpoint


    Examples
    --------
    The slightly shorter and more readable version:

    >>> @lazy_wrap
    ... def foo (data):
    ...     "Return first element"
    ...     return data[:3]
    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> data = foo(data)
    >>> data.realize()
    data_array([(8.6, 2), (53.36, 1), (nan, -1)],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@configuration@-@test@',
                      ['bar-fast', 'bar-slow', 'bar-failed'])])

    is equivalent to an explicit call to `lazy` on a user function:

    >>> def foo (data):
    ...     "Return first element"
    ...     return data[:3]
    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> data = lazy(foo, data)
    >>> data.realize()
    data_array([(8.6, 2), (53.36, 1), (nan, -1)],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@configuration@-@test@',
                      ['bar-fast', 'bar-slow', 'bar-failed'])])

    """
    @functools.wraps(func)
    def do_func(*args, **kwargs):
        return lazy(func, *args, **kwargs)
    return do_func


def lazy_wrap_realize(path=None):
    """Convenience decorator/wrapper for lazily-evaluated user functions.

    Returns the result of a call to `realize(path)` on the lazyly evaluated
    result.


    See also
    --------
    lazy_wrap


    Examples
    --------
    >>> @lazy_wrap_realize("/tmp/foo")
    ... def foo (data):
    ...     "Return first element"
    ...     return data[:3]
    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> foo(data)
    data_array([(8.6, 2), (53.36, 1), (nan, -1)],
               dtype=[('time', '<f8'), ('iterations', '<i8')],
               dims=[('@configuration@-@test@',
                      ['bar-fast', 'bar-slow', 'bar-failed'])])

    """
    def do_func1(func):
        @functools.wraps(func)
        def do_func2(*args, **kwargs):
            return lazy(func, *args, **kwargs).realize(path)
        return do_func2
    return do_func1


def lazy_wrap_checkpoint(path):
    """Convenience decorator/wrapper for lazily-evaluated user functions.

    Returns the result of a call to `checkpoint(path)` on the lazyly evaluated
    result.


    See also
    --------
    lazy_wrap


    Examples
    --------
    >>> @lazy_wrap_checkpoint("/tmp/foo")
    ... def foo (data):
    ...     "Return first element"
    ...     return data[:3]
    >>> data = lazy(extract_txt, "./data/maybe/@configuration@.csv",
    ...             fields_to_vars=["test"])
    >>> data = foo(data)

    """
    def do_func1(func):
        @functools.wraps(func)
        def do_func2(*args, **kwargs):
            return lazy(func, *args, **kwargs).checkpoint(path)
        return do_func2
    return do_func1


__all__ = [
    "find_files",
    "extract_func", "extract_txt", "extract_regex",
    "lazy", "lazy_wrap", "lazy_wrap_realize", "lazy_wrap_checkpoint",
]

__all__ += _errors.__all__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Parse and apply filter expressions."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2016, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import abc
import ply.lex as lex
import ply.yacc as yacc
import re
import six
import sys

from . import utils


class ParseError (AssertionError):
    """Exception signaling parse errors."""
    pass


class FilterError (Exception):
    """Exception signaling a filter with an unknown variable name."""

    def __init__(self, name, filter_):
        Exception.__init__(self, name, Filter(filter_))

    def __str__(self):
        return "Unknown variable '%s' in filter: %s" % (
            repr(self.args[0]), repr(str(self.args[1])))


class FilterExprError (Exception):
    """Exception signaling an unknown error during expression evaluation."""

    def __init__(self, expression, e):
        Exception.__init__(self, expression, e)

    def __str__(self):
        return "Error evaluating expression: %s\n  [was] %s: %s" %\
            (self.args[0],
             self.args[1].__class__.__name__,
             self.args[1])


class FilterEvalError (Exception):
    """Exception signaling an unknown error during filter evaluation."""

    def __init__(self, filter_, instance, e):
        Exception.__init__(self, filter_, instance, e)

    def __str__(self):
        return "Error evaluating filter '%s' with: %r\n  [was] %s: %s" %\
            (self.args[0], self.args[1],
             self.args[2].__class__.__name__,
             self.args[2])


class FilterUnallowedError (Exception):
    """Exception signaling a filter with an unallowed variable name."""

    def __init__(self, name, filter_):
        Exception.__init__(self, name, filter_)

    def __str__(self):
        return "Unallowed variable %r in filter: %s" % (self.args[0],
                                                        self.args[1])


class BaseChecker:
    """Abstract generic parser for expressions referencing Instance objects."""

    # AST elements

    class Atom (six.with_metaclass(abc.ABCMeta)):

        def __call__(self, *args):
            try:
                return self.evaluate(*args)
            except (KeyError, FilterExprError):
                raise
            except Exception as e:
                _, _, exc_traceback = sys.exc_info()
                new_e = FilterExprError(self, e)
                six.reraise(new_e.__class__, new_e, exc_traceback)

        @abc.abstractmethod
        def evaluate(self, *args):
            pass

        @abc.abstractmethod
        def copy(self):
            pass

        @abc.abstractmethod
        def validate(self, allowed):
            pass

        @abc.abstractmethod
        def __str__(self):
            pass

        @abc.abstractmethod
        def del_ids(self, dels):
            pass

    class AtomConstant (Atom):
        """Constant value."""
        def __init__(self, value):
            self._value = value

        def evaluate(self, *args):
            return self._value

        def copy(self):
            return BaseChecker.AtomConstant(self._value)

        def validate(self, allowed):
            pass

        def __str__(self):
            if isinstance(self._value, six.string_types):
                return "'%s'" % self._value
            else:
                return repr(self._value)

        def del_ids(self, dels):
            return self

    class AtomVar (Atom):
        """Variable reference."""
        def __init__(self, id):
            self._id = id

        def evaluate(self, var_dict):
            return var_dict[self._id]

        def copy(self):
            return BaseChecker.AtomVar(self._id)

        def validate(self, allowed):
            if self._id not in allowed:
                raise FilterUnallowedError(self._id, self)

        def __str__(self):
            return str(self._id)

        def del_ids(self, dels):
            if len(dels) == 0 or self._id in dels:
                del self
                return None
            return self

    class AtomVarExists (Atom):
        """Whether a variable exists."""
        def __init__(self, id):
            self._id = id

        def evaluate(self, var_dict):
            return self._id in var_dict

        def copy(self):
            return BaseChecker.AtomVarExists(self._id)

        def validate(self, allowed):
            if self._id not in allowed:
                raise FilterUnallowedError(self._id, self)

        def __str__(self):
            return "exists(%s)" % self._id

        def del_ids(self, dels):
            if len(dels) == 0 or self._id in dels:
                del self
                return None
            return self

    class AtomList (Atom):
        """List."""
        def __init__(self, items):
            self._items = items

        def evaluate(self, var_dict):
            return [e(var_dict) for e in self._items]

        def copy(self):
            return BaseChecker.AtomList([e.copy() for e in self._items])

        def validate(self, allowed):
            for e in self._items:
                e.validate(allowed)

        def __str__(self):
            return '[%s]' % ', '.join(str(e) for e in self._items)

        def del_ids(self, dels):
            items = [e.del_ids(dels) for e in self._items]
            items = [e for e in items if e is not None]
            if len(items) == 0:
                del self
                return None
            else:
                self._items = items
                return self

    class AtomBool (Atom):
        """Abstract boolean expression."""
        def __init__(self, e1, e2):
            self._e1 = e1
            self._e2 = e2

        def copy(self):
            return self.__class__(self._e1.copy(), self._e2.copy())

        def validate(self, allowed):
            self._e1.validate(allowed)
            self._e2.validate(allowed)

    class AtomCmp (AtomBool):
        """Abstract boolean comparison."""
        def del_ids(self, dels):
            self._e1 = self._e1.del_ids(dels)
            self._e2 = self._e2.del_ids(dels)
            if self._e1 is None or self._e2 is None:
                del self
                return None
            return self

    class AtomNEQ (AtomCmp):
        """Boolean 'not equal'."""
        def evaluate(self, *args):
            return self._e1(*args) != self._e2(*args)

        def __str__(self):
            return "%s != %s" % (self._e1, self._e2)

    class AtomEQ (AtomCmp):
        """Boolean 'equal'."""
        def evaluate(self, *args):
            return self._e1(*args) == self._e2(*args)

        def __str__(self):
            return "%s == %s" % (self._e1, self._e2)

    class AtomLT (AtomCmp):
        """Boolean 'less than'."""
        def evaluate(self, *args):
            return self._e1(*args) < self._e2(*args)

        def __str__(self):
            return "%s < %s" % (self._e1, self._e2)

    class AtomLE (AtomCmp):
        """Boolean 'less or equal'."""
        def evaluate(self, *args):
            return self._e1(*args) <= self._e2(*args)

        def __str__(self):
            return "%s <= %s" % (self._e1, self._e2)

    class AtomGT (AtomCmp):
        """Boolean 'greater than'."""
        def evaluate(self, *args):
            return self._e1(*args) > self._e2(*args)

        def __str__(self):
            return "%s > %s" % (self._e1, self._e2)

    class AtomGE (AtomCmp):
        """Boolean 'greater or equal'."""
        def evaluate(self, *args):
            return self._e1(*args) >= self._e2(*args)

        def __str__(self):
            return "%s >= %s" % (self._e1, self._e2)

    class AtomIN (AtomCmp):
        """Boolean 'in'."""
        def evaluate(self, *args):
            e1 = self._e1(*args)
            e2 = self._e2(*args)
            return e1 in e2

        def __str__(self):
            return "%s in %s" % (self._e1, self._e2)

    class AtomBoolOp (AtomBool):
        """Abstract boolean operation."""
        def del_ids(self, dels):
            self._e1 = self._e1.del_ids(dels)
            self._e2 = self._e2.del_ids(dels)
            if self._e1 is None:
                return self._e2
            elif self._e2 is None:
                return self._e1
            return self

    class AtomOR (AtomBoolOp):
        """Boolean 'or'."""
        def evaluate(self, *args):
            return self._e1(*args) or self._e2(*args)

        def __str__(self):
            res = ""
            if isinstance(self._e1, BaseChecker.AtomOR) or \
               isinstance(self._e1, BaseChecker.AtomCmp):
                res += "%s" % self._e1
            else:
                res += "( %s )" % self._e1
            res += " || "
            if isinstance(self._e2, BaseChecker.AtomOR) or \
               isinstance(self._e2, BaseChecker.AtomCmp):
                res += "%s" % self._e2
            else:
                res += "( %s )" % self._e2
            return res

    class AtomAND (AtomBoolOp):
        """Boolean 'and'."""
        def evaluate(self, *args):
            return self._e1(*args) and self._e2(*args)

        def __str__(self):
            res = ""
            if isinstance(self._e1,
                          (BaseChecker.AtomAND, BaseChecker.AtomCmp,
                           BaseChecker.AtomNOT)):
                res += "%s" % self._e1
            else:
                res += "( %s )" % self._e1
            res += " && "
            if isinstance(self._e2,
                          (BaseChecker.AtomAND, BaseChecker.AtomCmp,
                           BaseChecker.AtomNOT)):
                res += "%s" % self._e2
            else:
                res += "( %s )" % self._e2
            return res

    class AtomNOT (Atom):
        """Boolean 'not'."""
        def __init__(self, e):
            self._e = e

        def copy(self):
            return self.__class__(self._e.copy())

        def validate(self, allowed):
            self._e.validate(allowed)

        def del_ids(self, dels):
            self._e = self._e.del_ids(dels)
            if self._e is None:
                return None
            return self

        def evaluate(self, *args):
            return not self._e(*args)

        def __str__(self):
            if isinstance(self._e, (BaseChecker.AtomAND, BaseChecker.AtomOR)):
                return "! (" + str(self._e) + ")"
            else:
                return "! " + str(self._e)

    class AtomOp (Atom):
        """Abstract arithmetic operation."""
        def __init__(self, e1, e2):
            self._e1 = e1
            self._e2 = e2

        def copy(self):
            return self.__class__(self._e1.copy(), self._e2.copy())

        def validate(self, allowed):
            self._e1.validate(allowed)
            self._e2.validate(allowed)

        def del_ids(self, dels):
            self._e1 = self._e1.del_ids(dels)
            self._e2 = self._e2.del_ids(dels)
            if self._e1 is None or self._e2 is None:
                del self
                return None
            else:
                return self

    class AtomUMINUS (Atom):
        """Arithmetic negation."""
        def __init__(self, e):
            self._e = e

        def evaluate(self, *args):
            return - self._e(*args)

        def copy(self):
            return self.__class__(self._e.copy())

        def __str__(self):
            return "- %s" % self._e

        def validate(self, allowed):
            self._e.validate(allowed)

        def del_ids(self, dels):
            self._e = self._e.del_ids(dels)
            if self._e is None:
                del self
                return None
            else:
                return self

    class AtomPLUS (AtomOp):
        """Arithmetic sum."""
        def evaluate(self, *args):
            return self._e1(*args) + self._e2(*args)

        def __str__(self):
            return "%s + %s" % (self._e1, self._e2)

    class AtomMINUS (AtomOp):
        """Arithmetic substract."""
        def evaluate(self, *args):
            return self._e1(*args) - self._e2(*args)

        def __str__(self):
            return "%s - %s" % (self._e1, self._e2)

    class AtomTIMES (AtomOp):
        """Arithmetic multiply."""
        def evaluate(self, *args):
            return self._e1(*args) * self._e2(*args)

        def __str__(self):
            return "%s * %s" % (self._e1, self._e2)

    class AtomDIVIDE (AtomOp):
        """Arithmetic divide."""
        def evaluate(self, *args):
            return self._e1(*args) / self._e2(*args)

        def __str__(self):
            return "%s / %s" % (self._e1, self._e2)

    class AtomMOD (AtomOp):
        """Arithmetic modulus."""
        def evaluate(self, *args):
            return self._e1(*args) % self._e2(*args)

        def __str__(self):
            return "%s %% %s" % (self._e1, self._e2)

    @staticmethod
    def compile_regex(regex):
        try:
            return re.compile(str(regex))
        except re.error as e:
            _, _, exc_traceback = sys.exc_info()
            new_e = ParseError("Invalid regular expression: %s"
                               "\n  [was] %s: %s" % (regex,
                                                     e.__class__.__name__,
                                                     str(e)))
            six.reraise(new_e.__class__, new_e, exc_traceback)

    class AtomPBIND (AtomOp):
        """Pattern binding (regexp matching)."""
        def __init__(self, *args, **kwargs):
            BaseChecker.AtomOp.__init__(self, *args, **kwargs)
            self._cre2 = BaseChecker.compile_regex(self._e2())

        def evaluate(self, *args):
            e1 = str(self._e1(*args))
            m = self._cre2.match(e1)
            return m is not None

        def __str__(self):
            return "%s =~ %s" % (self._e1, self._e2)

    class AtomNPBIND (AtomOp):
        """Pattern binding (regexp matching)."""
        def __init__(self, *args, **kwargs):
            BaseChecker.AtomOp.__init__(self, *args, **kwargs)
            self._cre2 = BaseChecker.compile_regex(self._e2())

        def evaluate(self, *args):
            e1 = str(self._e1(*args))
            m = self._cre2.match(e1)
            return m is None

        def __str__(self):
            return "%s !~ %s" % (self._e1, self._e2)

    # Tokens
    reserved = {
        'True':   'TRUE',
        'False':  'FALSE',
        'in':     'IN',
        'not':    'NOT_STR',
        'exists': 'EXISTS',
    }

    tokens = [
        'STRING',
        'NUMBER',
        'LPAR', 'RPAR',
        'LSQB', 'RSQB', 'COMMA',
        'EQ', 'NEQ', 'LT', 'LE', 'GT', 'GE',
        'PLUS', 'MINUS', 'TIMES', 'DIVIDE', 'MOD',
        'AND', 'OR', 'NOT',
        'PBIND', 'NPBIND',
        'ID',
    ] + list(reserved.values())

    t_STRING = r"'[^']+'"

    def t_NUMBER(self, t):
        r'\d+\.?\d*'
        t.value = utils.str2num(t.value)
        assert not isinstance(t.value, str)
        return t

    t_LPAR = r'\('
    t_RPAR = r'\)'
    t_LSQB = r'\['
    t_RSQB = r'\]'
    t_COMMA = r','
    t_EQ = r'=='
    t_NEQ = r'!='
    t_LT = r'<'
    t_LE = r'<='
    t_GT = r'>'
    t_GE = r'>='
    t_PLUS = r'\+'
    t_MINUS = r'-'
    t_TIMES = r'\*'
    t_DIVIDE = r'/'
    t_MOD = r'%'
    t_AND = r'&&'
    t_OR = r'\|\|'
    t_NOT = r'!'
    t_PBIND = r'=~'
    t_NPBIND = r'!~'

    def t_ID(self, t):
        r'[a-zA-Z_][a-zA-Z0-9_]*'
        # check for reserved words
        t.type = self.reserved.get(t.value, 'ID')
        if t.type == "TRUE":
            t.value = True
        elif t.type == "FALSE":
            t.value = False
        return t

    # Ignored characters
    t_ignore = " \t"

    def t_error(self, t):
        raise ParseError("Illegal character in expression: %s" % t.value[0])

    # Parsing rules

    precedence = (
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES', 'DIVIDE'),
        ('left', 'MOD'),
        ('right', 'UMINUS'),
        ('left', 'AND', 'OR'),
        ('right', 'NOT'),
    )

    def p_bool(self, t):
        '''
        bool : TRUE
             | FALSE
        '''
        t[0] = self.AtomConstant(t[1])

    def p_bool_exists(self, t):
        '''
        bool : EXISTS LPAR ID RPAR
        '''
        t[0] = self.AtomVarExists(t[3])

    def p_bool_group(self, t):
        'bool : LPAR bool RPAR'
        t[0] = t[2]

    def p_bool_cmp(self, t):
        '''
        bool : expression cmp_op expression
        '''
        t[0] = t[2](t[1], t[3])
        if isinstance(t[1], self.AtomConstant) and \
           isinstance(t[3], self.AtomConstant):
            t[0] = self.AtomConstant(t[0]())

    def p_bool_cmp_op(self, t):
        '''
        cmp_op : EQ
               | NEQ
               | LT
               | LE
               | GT
               | GE
               | PBIND
               | NPBIND
        '''
        if t[1] == "==":
            t[0] = self.AtomEQ
        elif t[1] == "!=":
            t[0] = self.AtomNEQ
        elif t[1] == "<":
            t[0] = self.AtomLT
        elif t[1] == "<=":
            t[0] = self.AtomLE
        elif t[1] == ">":
            t[0] = self.AtomGT
        elif t[1] == ">=":
            t[0] = self.AtomGE
        elif t[1] == "=~":
            t[0] = self.AtomPBIND
        elif t[1] == "!~":
            t[0] = self.AtomNPBIND
        else:
            raise Exception

    def p_bool_cmp_multi(self, t):
        '''
        bool : expression cmp_op expression cmp_multi_rest
        '''
        left = t[2](t[1], t[3])
        cls, expression, rest = t[4]
        right = cls(t[3], expression)
        if rest is not None:
            right = self.AtomAND(right, rest)
        t[0] = self.AtomAND(left, right)

    def p_bool_cmp_multi_rest(self, t):
        '''
        cmp_multi_rest : cmp_op expression
                       | cmp_op expression cmp_multi_rest
        '''
        if len(t) == 3:
            t[0] = (t[1], t[2], None)
        elif len(t) == 4:
            cls, expression, rest = t[3]
            right = cls(t[2], expression)
            if rest is not None:
                right = self.AtomAND(right, rest)
            t[0] = (t[1], t[2], right)
        else:
            raise Exception

    def p_bool_in(self, t):
        '''
        bool : expression IN expression
        '''
        t[0] = self.AtomIN(t[1], t[3])

    def p_bool_not_in(self, t):
        '''
        bool : expression NOT_STR IN expression
        '''
        t[0] = self.AtomNOT(self.AtomIN(t[1], t[4]))

    def p_bool_bool(self, t):
        '''
        bool : bool OR  bool
             | bool AND bool
        '''
        if t[2] == "||":
            t[0] = self.AtomOR(t[1], t[3])
            if isinstance(t[1], self.AtomConstant):
                if t[1]():
                    t[0] = t[1]
                else:
                    t[0] = t[3]
        else:
            t[0] = self.AtomAND(t[1], t[3])
            if isinstance(t[1], self.AtomConstant):
                if t[1]():
                    t[0] = t[3]
                else:
                    t[0] = t[1]

    def p_bool_not(self, t):
        'bool : NOT bool %prec NOT'
        t[0] = self.AtomNOT(t[2])

    def p_bool_id(self, t):
        'bool : ID'
        t[0] = self.AtomVar(t[1])

    def p_expression_group(self, t):
        'expression : LPAR expression RPAR'
        t[0] = t[2]

    def p_expression_list(self, t):
        'expression : LSQB list RSQB'
        t[0] = self.AtomList(t[2])

    def p_expression_list_empty(self, t):
        'expression : LSQB RSQB'
        t[0] = self.AtomList([])

    def p_expression_list_full(self, t):
        '''
        list : expression
             | expression COMMA list
        '''
        if len(t) == 2:
            t[0] = [t[1]]
        elif len(t) == 4:
            t[0] = [t[1]] + t[3]
        else:
            raise Exception

    def p_expression_uminus(self, t):
        'expression : MINUS expression %prec UMINUS'
        t[0] = self.AtomUMINUS(t[2])
        if isinstance(t[2], self.AtomConstant):
            t[0] = self.AtomConstant(t[0]())

    def p_expression_plus(self, t):
        'expression : expression PLUS expression'
        t[0] = self.AtomPLUS(t[1], t[3])
        if isinstance(t[1], self.AtomConstant) and \
           isinstance(t[3], self.AtomConstant):
            t[0] = self.AtomConstant(t[0]())

    def p_expression_minus(self, t):
        'expression : expression MINUS expression'
        t[0] = self.AtomMINUS(t[1], t[3])
        if isinstance(t[1], self.AtomConstant) and \
           isinstance(t[3], self.AtomConstant):
            t[0] = self.AtomConstant(t[0]())

    def p_expression_times(self, t):
        'expression : expression TIMES expression'
        t[0] = self.AtomTIMES(t[1], t[3])
        if isinstance(t[1], self.AtomConstant) and \
           isinstance(t[3], self.AtomConstant):
            t[0] = self.AtomConstant(t[0]())

    def p_expression_divide(self, t):
        'expression : expression DIVIDE expression'
        t[0] = self.AtomDIVIDE(t[1], t[3])
        if isinstance(t[1], self.AtomConstant) and \
           isinstance(t[3], self.AtomConstant):
            t[0] = self.AtomConstant(t[0]())

    def p_expression_mod(self, t):
        'expression : expression MOD expression'
        t[0] = self.AtomMOD(t[1], t[3])
        if isinstance(t[1], self.AtomConstant) and \
           isinstance(t[3], self.AtomConstant):
            t[0] = self.AtomConstant(t[0]())

    def p_expression_number(self, t):
        '''
        expression : NUMBER
                   | TRUE
                   | FALSE
        '''
        t[0] = self.AtomConstant(t[1])

    def p_expression_str(self, t):
        'expression : STRING'
        t[0] = self.AtomConstant(t[1][1:-1])

    def p_expression_id(self, t):
        'expression : ID'
        t[0] = self.AtomVar(t[1])

    def p_error(self, t):
        if t:
            raise ParseError("Syntax error at '%s'" % t)
        else:
            raise ParseError("Unexpected end of filter expression")

    # Common interface

    # Each subclass must:
    # - define a '_checker' global attribute to hold the parser
    # - define a '_start' global attribute to hold the name of the starting
    #   rule
    # - the rule identified by '_start' must store the resulting AST into
    #   'self._result'

    @classmethod
    def _get(cls):
        """Get the parser."""
        if cls._checker is None:
            cls._checker = cls()
            cls._checker._build()
        return cls._checker

    def _build(self):
        """Build a new parser."""
        self._lexer = lex.lex(object=self, debug=0)
        self._parser = yacc.yacc(module=self, debug=0, write_tables=0,
                                 start=self._start)

    @classmethod
    def parse(cls, expression):
        """Parse 'expression'."""
        p = cls._get()
        try:
            p._parser.parse(expression, lexer=p._lexer, debug=0)
        except Exception as e:
            _, _, exc_traceback = sys.exc_info()
            new_e = ParseError("While parsing: %r"
                               "\n  [was] %s: %s" % (expression,
                                                     e.__class__.__name__,
                                                     str(e)))
            six.reraise(new_e.__class__, new_e, exc_traceback)
        return p._result


class _FilterChecker (BaseChecker):
    """Parser for Instance filters."""

    # AST elements

    class AtomFilter:
        def __init__(self, filter_):
            self._filter = filter_
            self._constant = self._is_constant()

        def __call__(self, instance):
            try:
                return self._filter(instance)
            except FilterExprError as e:
                _, _, exc_traceback = sys.exc_info()
                new_e = FilterEvalError(self, instance, e)
                six.reraise(new_e.__class__, new_e, exc_traceback)

        def copy(self):
            return _FilterChecker.AtomFilter(self._filter.copy())

        def __str__(self):
            return str(self._filter)

        def validate(self, allowed):
            self._filter.validate(allowed)

        def add(self, other):
            """Add a copy of 'other' into self."""
            self._filter = BaseChecker.AtomAND(self._filter,
                                               other._filter.copy())
            self._constant = self._is_constant()
            return self

        def _is_constant(self):
            return isinstance(self._filter, BaseChecker.AtomConstant)

        @property
        def constant(self):
            return self._constant

    # Parsing rules

    def p_filter(self, t):
        'filter : bool'
        self._result = _FilterChecker.AtomFilter(t[1])

    # Interface
    _checker = None
    _start = "filter"


def _filter_optimize(opt_dict, expression):
    """Try to optimize a filter into a `dict`."""
    if isinstance(expression, _FilterChecker.AtomFilter):
        return _filter_optimize(opt_dict, expression._filter)

    elif isinstance(expression, BaseChecker.AtomAND):
        return (_filter_optimize(opt_dict, expression._e1) and
                _filter_optimize(opt_dict, expression._e2))

    elif isinstance(expression, BaseChecker.AtomEQ):
        if (isinstance(expression._e1, BaseChecker.AtomVar) and
            isinstance(expression._e2, BaseChecker.AtomConstant)):
            if expression._e1._id in opt_dict:
                return opt_dict[expression._e1._id] == expression._e2._value
            opt_dict[expression._e1._id] = expression._e2._value
            return True
        elif (isinstance(expression._e1, BaseChecker.AtomConstant) and
              isinstance(expression._e2, BaseChecker.AtomVar)):
            if expression._e2._id in opt_dict:
                return opt_dict[expression._e2._id] == expression._e1._value
            opt_dict[expression._e2._id] = expression._e1._value
            return True
        else:
            return False

    else:
        return False


class Filter:
    """Boolean expression to check against a dict-like object.

    Any dict-like object can be checked against a filter to see if it
    positively matches the constraints in the filter, where variable references
    in the filter are used as keys to the target dict-like object.

    Filters are widely used to select specific contents of an instance group by
    selecting only those instances that match the filter.


    Parameters
    ----------
    filters : optional
        Filters to `AND` together.


    Raises
    ------
    ParseError
        Filter cannot be parsed (invalid syntax).


    Notes
    -----
    Elements in `filters` are `AND`'ed together into a single filter
    expression. Each element can take one of the following forms:

    * `str`: A textual filter that will be parsed.

    * `dict`: Shorthand where every ``key=value`` pair is considered a variable
      and value equality test.

    * `Filter`
    * `PFilter`

    Filters are transparently optimized for `AND`'ed expressions that check if
    a variable is equal to a given value.

    The available syntactical tokens are:

     ============================  ==========================
     Group expressions             ``( ... )``
     Lists of expressions          ``[ ... ]``
     Check membership              ``... in ...``, ``... not in ...``
     Reference other variables     VAR
     Check if a variable exists    ``exists(`` VAR ``)``
     Boolean values                ``True`` or ``False``
     Explicit numbers and strings  ``number`` or ``'string'``
     Simple math operators         ``+``, ``-``, ``*``, ``/``
     Comparisons                   ``<``, ``<=``, ``>``, ``>=``, ``==``, ``!=``
     Pattern binding               ``=~``, ``!~``
     Boolean operations            ``||``, ``&&``, ``!``
     ============================  ==========================

    In order to streamline the syntax of range checks, "multi-comparisons" are
    also available.

    Pattern binding operators match the left-hand argument with the right-hand
    regular expression. Both sides are converted into a string regardless of
    their type, and the the regular expression must match the beginning of the
    left-hand string.


    See also
    --------
    match


    Examples
    --------
    Filters are initially created by parsing strings, and can be composed using
    the sumation operator:

    >>> f = Filter("a < 3")
    >>> f & Filter("b == 4")
    Filter("a < 3 && b == 4")
    >>> f &= Filter("b == 4")
    >>> f
    Filter("a < 3 && b == 4")

    "Multi-comparisons" can be used as a shorthand for range checks:

    >>> Filter("1 < a < 3")
    Filter("1 < a && a < 3")
    >>> Filter("1 < a < b < 3")
    Filter("1 < a && a < b && b < 3")

    Once a filter has been "compiled", it can be later matched against
    :class:`dict`-like objects:

    >>> f.match(dict(a = 2, b = 4))
    True
    >>> f.match(dict(a = 3, b = 4))
    False

    Using dictionaries instead of strings makes ``f1`` and ``f2`` equivalent:

    >>> d = {"VAR1": 1, "VAR2": 2}
    >>> f1 = Filter(d)
    >>> f2 = Filter("VAR1 == 1 && VAR2 == 2")
    >>> str(f1) == str(f2)
    True
    >>> f1.as_dict() == d
    True

    Membership checks can simplify some filters:

    >>> Filter("VAR1 in [1, 3, 5]").match(dict(VAR1=3))
    True
    >>> Filter("VAR1 not in [1, 3, 5]").match(dict(VAR1=10))
    True

    """
    def __init__(self, *filters):
        self._expression = self._parse("True")
        self._dict = None

        if len(filters) == 0:
            return

        for f in filters:
            if isinstance(f, six.string_types):
                ff = self._parse(f)
                f = Filter()
                f._expression = ff
            elif isinstance(f, dict):
                ff = Filter()
                ff._dict = f
                f = ff
            elif isinstance(f, PFilter):
                f = f.realize()
            self._merge(f)

        self._optimize()

    @staticmethod
    def _parse(expression):
        return _FilterChecker.parse(expression)

    def _optimize(self):
        if self._expression.constant:
            return
        opt_dict = {}
        if _filter_optimize(opt_dict, self._expression):
            self._expression = self._parse("True")
            self._dict = opt_dict
        else:
            self._dict = None
        # sanity checks
        if self._dict is not None:
            assert self._expression.constant and self._expression({})

    def _merge(self, other):
        """Merge a Filter instance into this one."""
        if not isinstance(other, Filter):
            raise TypeError("invalid Filter type, found '%s'" % repr(other))

        if self._dict is None:
            self_expr = self._expression
        else:
            self_expr = self._parse(str(self))

        self_expr.add(self._parse(str(other)))
        self._expression = self._parse(str(self_expr))
        self._optimize()

    def __str__(self):
        """Return a string representation of the filter."""
        if self._dict is None:
            return str(self._expression)
        elif self._dict is not None:
            res = []
            for key, val in sorted(self._dict.items()):
                res.append("%s == %r" % (key, val))
            return " && ".join(res)

    def __repr__(self):
        return "Filter(\"%s\")" % str(self)

    def validate(self, allowed):
        """Validate that variables in the filter are present in the given set.

        Parameters
        ----------
        allowed : set of variable names
            Set of variable names to allow on the filter.


        Raises
        ------
        FilterUnallowedError
            Filter contains an unallowed variable.

        """
        if self._dict is not None:
            for var in self._dict.keys():
                if var not in allowed:
                    raise FilterUnallowedError(var, self)
        else:
            try:
                self._expression.validate(allowed)
            except FilterUnallowedError as e:
                raise FilterUnallowedError(e.args[0], self)

    def as_dict(self):
        """Return a :class:`dict` representation of this filter.

        Returns
        -------
        dict or None : The optimized dict representation of this filter.

        Notes
        -----
        If there is no optimized dict representation available, returns
        ``None``.

        """
        if self._dict is not None:
            return dict(self._dict)
        else:
            return None

    def match(self, source):
        """Check if the given :class:`dict` matches this filter.

        Parameters
        ----------
        source : dict-like
            Dictionary to match this filter against.

        Returns
        -------
        bool : Whether the match is positive or not.

        Raises
        ------
        FilterError
            Filter contains a variable name not present in `source`.
        FilterEvalError
            An error happened while evaluating the filter.

        See also
        --------
        validate

        """
        if self._dict is None and self._expression.constant:
            return self._expression({})
        assert isinstance(source, dict)

        try:
            if self._expression is not None:
                if not self._expression(source):
                    return False

            if self._dict is not None:
                for key, val in six.iteritems(self._dict):
                    if source[key] != val:
                        return False

        except KeyError as e:
            raise FilterError(e.args[0], self)

        return True

    @utils.deprecated
    def __iadd__(self, other):
        """Add an `AND`'ed filter object."""
        return self.__iand__(other)

    @utils.deprecated
    def __add__(self, other):
        """Add two `AND`'ed filters together into a new one."""
        return self.__and__(other)

    def __iand__(self, other):
        if not isinstance(other, Filter):
            other = Filter(other)
        self._merge(other)
        return self

    def __and__(self, other):
        res = Filter(self)
        res &= other
        return res

    def __or__(self, other):
        return Filter("(%s) || (%s)" % (self, other))


class PFilter (object):
    """Programmatic version of `Filter`.

    Provides a programmatic interface to constructing filters.

    Notes
    -----
    Unless noted, it supports the same operators as regular `Filter` instances.

    Boolean operations ``||``, ``&&`` and ``!`` must be written as ``|``, ``&``
    and ``~`` (respectively); just be sure to use parentheses around the groups
    you are and'ing or or'ing together . Pattern binding is not supported.


    Examples
    --------
    Filters constructed from strings or programmatically are equivalent (just
    remember to use parentheses):

    >>> Filter("!a || b % 2 == 0")
    Filter("( ! a ) || b % 2 == 0")
    >>> v_ = PFilter()
    >>> (~ v_.a) | (v_.b % 2 == 0)
    Filter("( ! a ) || b % 2 == 0")

    Programmatic filters can also be combined with other filters:

    >>> (v_.c == "foo") | Filter("b == 2")
    Filter("c == 'foo' || b == 2")

    An actual `Filter` instance can be obtained with the `realize` method:

    >>> isinstance(v_.a == 1, Filter)
    False
    >>> isinstance((v_.a == 1).realize(), Filter)
    True

    """
    def __init__(self, left=None):
        if left is None:
            left = ""
        else:
            left = str(left)
        self._other = left

    def __str__(self):
        return str(self.realize())

    def __repr__(self):
        return repr(self.realize())

    def realize(self):
        return Filter(self._other)

    def _construct(self, other):
        res = PFilter()
        res._other = other
        return res

    def __getattr__(self, name):
        if self._other == "":
            return self._construct(name)
        else:
            return self._construct("%s %s" % (self._other, name))

    # Generic operation handler

    def _operate(self, op, other, as_filter=False):
        if isinstance(other, PFilter):
            res_fmt = "(%s %s %s)"
            return self._construct(res_fmt % (self._other, op, other._other))
        elif isinstance(other, six.string_types) and not as_filter:
            res_fmt = "(%s %s '%s')"
            return self._construct(res_fmt % (self._other, op, other))
        else:
            res_fmt = "(%s %s %s)"
            return self._construct(res_fmt % (self._other, op, other))

    # Arithmetic

    def __neg__(self, other):
        return self._operate("-", other)

    def __add__(self, other):
        return self._operate("+", other)

    def __sub__(self, other):
        return self._operate("-", other)

    def __mul__(self, other):
        return self._operate("*", other)

    def __div__(self, other):
        return self._operate("/", other)

    def __mod__(self, other):
        return self._operate("%", other)

    # Comparison

    def __lt__(self, other):
        return self._operate("<", other)

    def __le__(self, other):
        return self._operate("<=", other)

    def __gt__(self, other):
        return self._operate(">", other)

    def __ge__(self, other):
        return self._operate(">=", other)

    def __eq__(self, other):
        return self._operate("==", other)

    def __ne__(self, other):
        return self._operate("!=", other)

    # Boolean

    def __and__(self, other):
        return self._operate("&&", other, as_filter=True)

    def __or__(self, other):
        return self._operate("||", other, as_filter=True)

    def __invert__(self):
        return self._construct("!(%s)" % self._other)
